-- 
-- CPU Performance Diagnostic T-SQL script library. Kevin Kline - kevin.kline@solarwinds.com,
-- 	@kekline on social media, blog - https://thwack.solarwinds.com/members/kekline/blogs


-- STEP 1 - What is making the users wait on our SQL Server? Is CPU even one of the reasons?
-- First, let's check signal vs resource waits at the instance level
-- Signal Waits above 10-15% is usually a sign of CPU pressure
SELECT
  CAST(100.0 * SUM(signal_wait_time_ms) / SUM(wait_time_ms) AS NUMERIC(20, 2))
    AS [%signal (cpu) waits],
  CAST(
    100.0 * SUM(wait_time_ms - signal_wait_time_ms) / SUM(wait_time_ms) AS NUMERIC(20, 2))
    AS [%resource waits]
FROM
  sys.dm_os_wait_stats;

-- STEP 1 - What is making the users wait on our SQL Server? Is memory even one of the reasons?
--
-- Original author - Paul Randal, @PaulRandal, http://www.sqlskills.com/blogs/paul/wait-statistics-or-please-tell-me-where-it-hurts/
-- 	The most important wait stats for CPU are: SOS_Scheduler_Yield, CXPacket, .

WITH [Waits] AS
    (SELECT
        [wait_type], [wait_time_ms] / 1000.0 AS [WaitS],
        ([wait_time_ms] - [signal_wait_time_ms]) / 1000.0 AS [ResourceS],
        [signal_wait_time_ms] / 1000.0 AS [SignalS],
        [waiting_tasks_count] AS [WaitCount],
        100.0 * [wait_time_ms] / SUM ([wait_time_ms]) OVER() AS [Percentage],
        ROW_NUMBER() OVER(ORDER BY [wait_time_ms] DESC) AS [RowNum]
    FROM sys.dm_os_wait_stats
    WHERE [wait_type] NOT IN (
        N'BROKER_EVENTHANDLER',             N'BROKER_RECEIVE_WAITFOR',
        N'BROKER_TASK_STOP',                N'BROKER_TO_FLUSH',
        N'BROKER_TRANSMITTER',              N'CHECKPOINT_QUEUE',
        N'CHKPT',                           N'CLR_AUTO_EVENT',
        N'CLR_MANUAL_EVENT',                N'CLR_SEMAPHORE',
        N'DBMIRROR_DBM_EVENT',              N'DBMIRROR_EVENTS_QUEUE',
        N'DBMIRROR_WORKER_QUEUE',           N'DBMIRRORING_CMD',
        N'DIRTY_PAGE_POLL',                 N'DISPATCHER_QUEUE_SEMAPHORE',
        N'EXECSYNC',                        N'FSAGENT',
        N'FT_IFTS_SCHEDULER_IDLE_WAIT',     N'FT_IFTSHC_MUTEX',
        N'HADR_CLUSAPI_CALL',               N'HADR_FILESTREAM_IOMGR_IOCOMPLETION',
        N'HADR_LOGCAPTURE_WAIT',            N'HADR_NOTIFICATION_DEQUEUE',
        N'HADR_TIMER_TASK',                 N'HADR_WORK_QUEUE',
        N'KSOURCE_WAKEUP',                  N'LAZYWRITER_SLEEP',
        N'LOGMGR_QUEUE',                    N'ONDEMAND_TASK_QUEUE',
        N'PWAIT_ALL_COMPONENTS_INITIALIZED',
        N'QDS_PERSIST_TASK_MAIN_LOOP_SLEEP',
        N'QDS_CLEANUP_STALE_QUERIES_TASK_MAIN_LOOP_SLEEP',
        N'REQUEST_FOR_DEADLOCK_SEARCH',     N'RESOURCE_QUEUE',
        N'SERVER_IDLE_CHECK',               N'SLEEP_BPOOL_FLUSH',
        N'SLEEP_DBSTARTUP',                 N'SLEEP_DCOMSTARTUP',
        N'SLEEP_MASTERDBREADY',             N'SLEEP_MASTERMDREADY',
        N'SLEEP_MASTERUPGRADED',            N'SLEEP_MSDBSTARTUP',
        N'SLEEP_SYSTEMTASK',                N'SLEEP_TASK',
        N'SLEEP_TEMPDBSTARTUP',             N'SNI_HTTP_ACCEPT',
        N'SP_SERVER_DIAGNOSTICS_SLEEP',     N'SQLTRACE_BUFFER_FLUSH',
        N'SQLTRACE_INCREMENTAL_FLUSH_SLEEP',
        N'SQLTRACE_WAIT_ENTRIES',           N'WAIT_FOR_RESULTS',
        N'WAITFOR',                         N'WAITFOR_TASKSHUTDOWN',
        N'WAIT_XTP_HOST_WAIT',              N'WAIT_XTP_OFFLINE_CKPT_NEW_LOG',
        N'WAIT_XTP_CKPT_CLOSE',             N'XE_DISPATCHER_JOIN',
        N'XE_DISPATCHER_WAIT',              N'XE_TIMER_EVENT')
    AND [waiting_tasks_count] > 0
 )
SELECT
    MAX ([W1].[wait_type]) AS [WaitType],
    CAST (MAX ([W1].[WaitS]) AS DECIMAL (16,2)) AS [Wait_S],
    CAST (MAX ([W1].[ResourceS]) AS DECIMAL (16,2)) AS [Resource_S],
    CAST (MAX ([W1].[SignalS]) AS DECIMAL (16,2)) AS [Signal_S],
    MAX ([W1].[WaitCount]) AS [WaitCount],
    CAST (MAX ([W1].[Percentage]) AS DECIMAL (5,2)) AS [Percentage],
    CAST ((MAX ([W1].[WaitS]) / MAX ([W1].[WaitCount])) AS DECIMAL (16,4)) AS [AvgWait_S],
    CAST ((MAX ([W1].[ResourceS]) / MAX ([W1].[WaitCount])) AS DECIMAL (16,4)) AS [AvgRes_S],
    CAST ((MAX ([W1].[SignalS]) / MAX ([W1].[WaitCount])) AS DECIMAL (16,4)) AS [AvgSig_S]
FROM [Waits] AS [W1]
	INNER JOIN [Waits] AS [W2] ON [W2].[RowNum] <= [W1].[RowNum]
GROUP BY [W1].[RowNum]
	-- HAVING SUM ([W2].[Percentage]) - MAX ([W1].[Percentage]) < 95; -- percentage threshold
GO

-- Step 2 - Check out CPU configuration
-- Hardware information, Original author: Glenn Berry at https://glennsqlperformance.com/
-- (Cannot distinguish between HT and multi-core)
SELECT cpu_count AS [Logical CPU Count], scheduler_count, hyperthread_ratio AS [Hyperthread Ratio],
	cpu_count/hyperthread_ratio AS [Physical CPU Count], 
	physical_memory_kb/1024 AS [Physical Memory (MB)], committed_kb/1024 AS [Committed Memory (MB)],
	committed_target_kb/1024 AS [Committed Target Memory (MB)],
	max_workers_count AS [Max Workers Count], affinity_type_desc AS [Affinity Type], 
	sqlserver_start_time AS [SQL Server Start Time], virtual_machine_type_desc AS [Virtual Machine Type]  
FROM sys.dm_os_sys_info WITH (NOLOCK) OPTION (RECOMPILE);

-- Get information on location, time and size of any memory dumps from SQL Server 
--    This will not return any rows if you have 
--    not had any processor dumps (which is a good thing)
SELECT [filename], creation_time, size_in_bytes/1048576.0 AS [Size (MB)]
FROM sys.dm_server_memory_dumps WITH (NOLOCK) 
ORDER BY creation_time DESC OPTION (RECOMPILE);

-- Original author: Joe Sack at Microsoft
--   Configuration CPU settings
SELECT name, value, value_in_use, is_advanced,is_dynamic
FROM sys.configurations 
WHERE name IN
	('affinity mask',
	'affinity64 mask',
	'cost threshold for parallelism',
	'lightweight pooling',
	'max degree of parallelism',
	'max worker threads',
	'priority boost')
ORDER BY name;
GO


-- Step 3 - Learn about the server CPU consumption by database.
-- Get CPU utilization by database (CPU Usage by Database)
--    However, this only reflects CPU usage from the currently cached query plans
WITH DB_CPU_Stats
AS
(SELECT pa.DatabaseID, DB_Name(pa.DatabaseID) AS [Database Name], SUM(qs.total_worker_time/1000) AS [CPU_Time_Ms]
 FROM sys.dm_exec_query_stats AS qs WITH (NOLOCK)
 CROSS APPLY (SELECT CONVERT(int, value) AS [DatabaseID] 
              FROM sys.dm_exec_plan_attributes(qs.plan_handle)
              WHERE attribute = N'dbid') AS pa
 GROUP BY DatabaseID)
SELECT ROW_NUMBER() OVER(ORDER BY [CPU_Time_Ms] DESC) AS [CPU Rank],
       [Database Name], [CPU_Time_Ms] AS [CPU Time (ms)], 
       CAST([CPU_Time_Ms] * 1.0 / SUM([CPU_Time_Ms]) OVER() * 100.0 AS DECIMAL(5, 2)) AS [CPU Percent]
FROM DB_CPU_Stats
WHERE DatabaseID <> 32767 -- ResourceDB
ORDER BY [CPU Rank] OPTION (RECOMPILE);

-- Get Average Task Counts (run multiple times)  (Avg Task Counts)
-- Sustained values above 10 suggest further investigation in that area
-- High Avg Task Counts are often caused by blocking/deadlocking or other resource contention
-- Sustained values above 1 suggest further investigation in that area
-- High Avg Runnable Task Counts are a good sign of CPU pressure
SELECT AVG(current_tasks_count) AS [Avg Task Count], 
	AVG(work_queue_count) AS [Avg Work Queue Count],
	AVG(runnable_tasks_count) AS [Avg Runnable Task Count],
	AVG(pending_disk_io_count) AS [Avg Pending DiskIO Count]
FROM sys.dm_os_schedulers WITH (NOLOCK)
WHERE scheduler_id < 255 OPTION (RECOMPILE);
GO


-- Get CPU Utilization History for last 256 minutes (in one minute intervals) by SPID (CPU Utilization History)
-- Look at the trend over the entire period 
-- Also look at high sustained 'Other Process' CPU Utilization values
-- Note: This query sometimes gives inaccurate results (negative values)
-- on high core count (> 64 cores) systems
DECLARE @ts_now bigint = (SELECT cpu_ticks/(cpu_ticks/ms_ticks) FROM sys.dm_os_sys_info WITH (NOLOCK)); 

SELECT TOP(256) SQLProcessUtilization AS [SQL Server Process CPU Utilization], 
               SystemIdle AS [System Idle Process], 
               100 - SystemIdle - SQLProcessUtilization AS [Other Process CPU Utilization], 
               DATEADD(ms, -1 * (@ts_now - [timestamp]), GETDATE()) AS [Event Time] 
FROM (SELECT record.value('(./Record/@id)[1]', 'int') AS record_id, 
			record.value('(./Record/SchedulerMonitorEvent/SystemHealth/SystemIdle)[1]', 'int') 
			AS [SystemIdle], 
			record.value('(./Record/SchedulerMonitorEvent/SystemHealth/ProcessUtilization)[1]', 'int') 
			AS [SQLProcessUtilization], [timestamp] 
	  FROM (SELECT [timestamp], CONVERT(xml, record) AS [record] 
			FROM sys.dm_os_ring_buffers WITH (NOLOCK)
			WHERE ring_buffer_type = N'RING_BUFFER_SCHEDULER_MONITOR' 
			AND record LIKE N'%<SystemHealth>%') AS x) AS y 
ORDER BY record_id DESC OPTION (RECOMPILE);
GO


-- Step 4 - What are the bad queries for CPU consumption?
-- Get top total worker time queries for entire instance (Top Worker Time Queries)
-- This query is also helpful for troubleshooting blocking and deadlocking issues
SELECT TOP(50) DB_NAME(t.[dbid]) AS [Database Name], 
	REPLACE(REPLACE(LEFT(t.[text], 255), CHAR(10),''), CHAR(13),'') AS [Short Query Text],  
	qs.total_worker_time AS [Total Worker Time], qs.min_worker_time AS [Min Worker Time],
	qs.total_worker_time/qs.execution_count AS [Avg Worker Time], 
	qs.max_worker_time AS [Max Worker Time], 
	qs.min_elapsed_time AS [Min Elapsed Time], 
	qs.total_elapsed_time/qs.execution_count AS [Avg Elapsed Time], 
	qs.max_elapsed_time AS [Max Elapsed Time],
	qs.min_logical_reads AS [Min Logical Reads],
	qs.total_logical_reads/qs.execution_count AS [Avg Logical Reads],
	qs.max_logical_reads AS [Max Logical Reads], 
	qs.execution_count AS [Execution Count],
	CASE WHEN CONVERT(nvarchar(max), qp.query_plan) COLLATE Latin1_General_BIN2 LIKE N'%<MissingIndexes>%' THEN 1 ELSE 0 END AS [Has Missing Index],
	qs.creation_time AS [Creation Time]
	--,t.[text] AS [Query Text], qp.query_plan AS [Query Plan] -- uncomment out these columns if not copying results to Excel
FROM sys.dm_exec_query_stats AS qs WITH (NOLOCK)
	CROSS APPLY sys.dm_exec_sql_text(plan_handle) AS t 
	CROSS APPLY sys.dm_exec_query_plan(plan_handle) AS qp 
ORDER BY qs.total_worker_time DESC OPTION (RECOMPILE);


-- Get input buffer information for the current database (Input Buffer)
SELECT es.session_id, DB_NAME(es.database_id) AS [Database Name],
	   es.login_time, es.cpu_time, es.logical_reads,
	   es.[status], ib.event_info AS [Input Buffer]
FROM sys.dm_exec_sessions AS es WITH (NOLOCK)
CROSS APPLY sys.dm_exec_input_buffer(es.session_id, NULL) AS ib
WHERE es.database_id = DB_ID()
AND es.session_id > 50
AND es.session_id <> @@SPID OPTION (RECOMPILE);


-- Could over-aggressive parallelism be the CPU resource hog?
-- Original author: Jonathan Kehayias, https://www.sqlskills.com/blogs/jonathan/tuning-cost-threshold-for-parallelism-from-the-plan-cache/
SET TRANSACTION ISOLATION LEVEL READ UNCOMMITTED; 
WITH XMLNAMESPACES   
(DEFAULT 'http://schemas.microsoft.com/sqlserver/2004/07/showplan')  
SELECT  
     query_plan AS CompleteQueryPlan, 
     n.value('(@StatementText)[1]', 'VARCHAR(4000)') AS StatementText, 
     n.value('(@StatementOptmLevel)[1]', 'VARCHAR(25)') AS StatementOptimizationLevel, 
     n.value('(@StatementSubTreeCost)[1]', 'VARCHAR(128)') AS StatementSubTreeCost, 
     n.query('.') AS ParallelSubTreeXML,  
     ecp.usecounts, 
     ecp.size_in_bytes 
FROM sys.dm_exec_cached_plans AS ecp 
CROSS APPLY sys.dm_exec_query_plan(plan_handle) AS eqp 
CROSS APPLY query_plan.nodes('/ShowPlanXML/BatchSequence/Batch/Statements/StmtSimple') AS qn(n) 
WHERE  n.query('.').exist('//RelOp[@PhysicalOp="Parallelism"]') = 1;


-- CPU Ring Buffers and Query Store
SELECT top 10 id,
           SQLServerCPUUtilization,
           100 - SystemIdle - SQLServerCPUUtilization AS NonSQLCPUUtilization -- SystemIdle on Linux will be 0
FROM
  (SELECT record.value('(./Record/@id)[1]', 'int') AS id,
          record.value('(./Record/SchedulerMonitorEvent/SystemHealth/SystemIdle)[1]', 'int') AS SystemIdle,
          record.value('(./Record/SchedulerMonitorEvent/SystemHealth/ProcessUtilization)[1]', 'int') AS SQLServerCPUUtilization,
          TIMESTAMP
   FROM
     (SELECT TIMESTAMP,
             convert(XML, record) AS record
      FROM sys.dm_os_ring_buffers
      WHERE ring_buffer_type = N'RING_BUFFER_SCHEDULER_MONITOR'
        AND record like '%<SystemHealth>%') AS RingBufferInfo) AS TabularInfo
ORDER BY id DESC;


-- Step 4 - correlate the wait stat information with PerfMon counters indicating memory pressure 
-- See stored procedure Sp_PerformanceCounters...


-- Step 5 - correlate symptoms with further metrics from important DMVs 
-- Usually only needed in long-term, hard to find situatiions
-- Get CPU consumption by SQLOS scheduler
SELECT	*
FROM sys.dm_os_schedulers
WHERE status LIKE 'VISIBLE ONLINE%';

-- Get schedulers over time
SELECT 
    EventTime,
    n.value('(SystemIdle)[1]', 'int') AS idle_cpu,
    100-(n.value('(SystemIdle)[1]', 'int') +
			n.value('(ProcessUtilization)[1]', 'int')) AS nonsql_cpu,
    n.value('(ProcessUtilization)[1]', 'int') AS sql_cpu
FROM (SELECT
		DATEADD (ss, (-1 * ((cpu_ticks / CONVERT (float, ( cpu_ticks / ms_ticks ))) - [timestamp])/1000), GETDATE()) AS EventTime, 
		CONVERT (xml, record) AS record
	FROM sys.dm_os_ring_buffers 
	CROSS JOIN sys.dm_os_sys_info
	WHERE ring_buffer_type = 'RING_BUFFER_SCHEDULER_MONITOR') AS t
CROSS APPLY record.nodes('/Record/SchedulerMonitorEvent/SystemHealth') AS x(n)
ORDER BY EventTime DESC;

-- Compare to PID in Windows Task Manager
SELECT SERVERPROPERTY('processid') AS [processid];


-- APPENDIX
-- The ultra-query to find top consumer queries by a variety of metrics.
-- Original Author: Jimmy May, Sandisk, @aspiringgeek
SELECT   TOP 25
  -- the following four columns are NULL for ad hoc and prepared batches
  DB_Name(qp.dbid) AS dbname,
  qp.dbid,
  qp.objectid,
  qp.number, 
  -- qp.query_plan, --the query plan can be *very* useful; enable if desired
  qt.text,
  SUBSTRING(qt.text, (qs.statement_start_offset / 2) + 1, 
		((CASE statement_end_offset WHEN -1 THEN DATALENGTH(qt.text) 
			ELSE qs.statement_end_offset 
		END - qs.statement_start_offset) / 2) + 1) AS statement_text,
  qs.creation_time,
  qs.last_execution_time,
  qs.execution_count,
  qs.total_worker_time / qs.execution_count AS avg_worker_time,
  qs.total_physical_reads / qs.execution_count AS avg_physical_reads,
  qs.total_logical_reads / qs.execution_count AS avg_logical_reads,
  qs.total_logical_writes / qs.execution_count AS avg_logical_writes,
  qs.total_elapsed_time / qs.execution_count AS avg_elapsed_time,
  --qs.total_clr_time / qs.execution_count AS avg_clr_time,
  qs.total_worker_time,
  qs.last_worker_time,
  qs.min_worker_time,
  qs.max_worker_time,
  qs.total_physical_reads,
  qs.last_physical_reads,
  qs.min_physical_reads,
  qs.max_physical_reads,
  qs.total_logical_reads,
  qs.last_logical_reads,
  qs.min_logical_reads,
  qs.max_logical_reads,
  qs.total_logical_writes,
  qs.last_logical_writes,
  qs.min_logical_writes,
  qs.max_logical_writes,
  qs.total_elapsed_time,
  qs.last_elapsed_time,
  qs.min_elapsed_time,
  qs.max_elapsed_time,
  --qs.total_clr_time, qs.last_clr_time, qs.min_clr_time, qs.max_clr_time, 
  --, qs.sql_handle , qs.statement_start_offset , qs.statement_end_offset
  qs.plan_generation_num -- , qp.encrypted
FROM sys.dm_exec_query_stats AS qs
  CROSS APPLY sys.dm_exec_query_plan(qs.plan_handle) AS qp
  CROSS APPLY sys.dm_exec_sql_text(qs.sql_handle) AS qt
-- WHERE...
ORDER BY qs.total_logical_reads  DESC  --Reads

--ORDER BY qs.execution_count      DESC  --Frequency
--ORDER BY qs.total_elapsed_time   DESC  --Durn
--ORDER BY qs.total_logical_reads  DESC  --Reads
--ORDER BY qs.total_logical_writes DESC  --Writes
--ORDER BY qs.total_physical_reads DESC  --PhysicalReads
--ORDER BY avg_worker_time         DESC  --AvgCPU
--ORDER BY avg_elapsed_time        DESC  --AvgDurn
--ORDER BY avg_logical_reads       DESC  --AvgReads
--ORDER BY avg_logical_writes      DESC  --AvgWrites
--ORDER BY avg_physical_reads      DESC  --AvgPhysicalReads

--Sample WHERE clauses. Modify WHERE & ORDER BY clauses to suit circumstances
--WHERE last_execution_time > '20070507 15:00'
--WHERE execution_count = 1
--WHERE SUBSTRING(qt.text, (qs.statement_start_offset/2) + 1,
--    ((CASE statement_end_offset
--        WHEN -1 THEN DATALENGTH(qt.text)
--        ELSE qs.statement_end_offset END
--            - qs.statement_start_offset)/2) + 1)
--      LIKE '%MyText%'

